!*****************************************************
!this module contains all elements that are used in 
!computation and that depend on the parameters vector
!*****************************************************


!******************
!Marina Pavan
!May 25 2005
!******************


module mod_utils

use mod_types
use mod_computation


implicit none

!*****************************
!variables:
!*****************************



real(dp), dimension(ny*nz)::zy  !income
real(dp), dimension(nny,nny)::pyy  !Markov transition matrix for y
real(dp), dimension(nny)::py  !unconditional probabilities

real(dp), dimension(max_age)::eta !age specific productivity 
real(dp), dimension(max_age):: ybar  !pension


real(dp), dimension(nybz,nybz)::pbzyy  !transition matrix of income/betas
integer, dimension(nybz)::index_Azy,index_z,index_y,index_Ags  !indexes coresponding to each iyb in pyb (to recognize proper state)
real(dp),dimension(nybz)::pstat  !joint stationary distribution over zy and b
real(dp), dimension(nbeta)::pbetastat  !stationary distribution for betas

real(dp),dimension(max_age-ret_age+1,ns,nk)::exp_pens  !expected future stream of pensions (discounted)
real(dp),dimension(ret_age,ny,nk)::exp_inc  !expected future stream of income + pensions (discounted)
real(dp), dimension(ny,ny)::payay  !markov transition matrix for the whole y*Ags


!real(dp),dimension(nybz/ns,nybz/ns)::pbzyy_sim  !for simulation in random_gen1
!real(dp),dimension(nybz/ns)::pstat_sim

!real(dp)::coh(ns,nk,ncoh)  !cash-on-hand
!integer,dimension(nh,nb,ns,nk)::coh_index  !indicates the position in coh of any triple (ih,id,ib)
!integer::h_index(ncoh)
!integer,dimension(2,nh,nb,ns,nk)::indsnext
!real(dp),dimension(2,nh,nb,ns,nk)::valsnext

!value function:
real(dp)::v_f(max_age,fulnh,fulnb,nybz,nk)
!policy functions:
integer::pol_h(max_age,fulnh,fulnb,nybz,nk),pol_b(max_age,fulnh,fulnb,nybz,nk) !policy functions 
!integer::ipol_h(nh,nb,nybz,nk,nhh),ipol_b(nh,nb,nybz,nk,nhh) !policy functions (indexes)

!real(dp),allocatable::DISVEC(:)

!integer,dimension(ns,nk)::index_zero  !index of zero cash-on-hand 

contains

!***************************************************************************
!subroutine markovappr approximates a first-order autoregressive
!process with a Markov chain
!
!y_t=y_rho*y_(t-1)+u_t
!
!u_t is a Gaussian white noise process with standard deviation
!u_sigma.
!
!width_y determines the width of discretized state space. Tauchen uses 
!width_y=3, ymax=width_y*var(y),ymin=-width_y*var(y), ymin and ymax are two boundary points
!
!ny is the number of possible states chosen to approximate the y_t process
!usually ny=9 should be fine - but for the moment ny=2
!
!pyy is the transition matrix of the Markov chain
!y is the discretized state space of y_t
!
!arho is the theoretical first order autoregression coefficient for
!Markov chain
!asigma is the theoretical standard deviation for Markov chain y_t
!
!****************************************************************************

subroutine  markovappr

use mod_types
use mod_param
use mod_computation
!use NUMERICAL_LIBRARIES

implicit none

!***********************
!variables:
!***********************
real(dp), dimension(nny)::y     !persistent part of earnings*aggregate shock
real(dp), dimension(ns)::ya
real(dp), dimension(nz)::z   !fixed effect of income
!real(dp), dimension(ny,ny)::payay  !markov transition matrix for the whole y*Ags
real(dp)::exp_y,exp_eta  !expected earnings
real(dp),dimension(ny)::Ay
real(dp), dimension(nz)::pz   !distribution of fixed effect
real(dp)::zmin,zmax,we
integer::iz

integer, parameter::nyb=ny*nbeta

real(dp)::u_sigma,u_sigmaa  !stdev of White Noise
real(dp)::var_y     !var of y
real(dp)::ymin,ymax,yamin,yamax
real(dp)::w  !length of interval

real(dp),dimension(nny,nny)::transpyy  !to compute stationary distributions
real(dp),dimension(nny)::py1

real(dp),dimension(nbeta,nbeta)::transpb
real(dp),dimension(nbeta)::pb1

real(dp), dimension(nyb)::pby  !joint prob. for y and betas
real(dp), dimension(nyb,nyb)::pbyy  !transition prob. for y and betas
real(dp), dimension(nny*nbeta,nny*nbeta)::pbyy_sim

integer, dimension(nz,ny)::num

integer::iy1,iy2,m,n,ib,ibprime,iy,iyprime,is,iss,age
real(dp)::test,epsi_trans

integer, dimension(ret_age-1)::m1,m2

real(dp),dimension(nny),parameter::Ivect=1.0_dp
real(dp),dimension(nbeta), parameter::Ivect1=1.0_dp

real(dp)::totlabor

real(8) :: dnordf

external dnordf


!*****************************************************
!assign age specific productivity
!taking into account that possibly pery>1
!while original calibration was for pery=1
!*****************************************************

m2=0
m1=0
do m=1,ret_age-1
	m1(m)=(m-1)*pery+1  !if m=1,m1=1; if m=2,m1=4, and so on (m=1 represents first three ages in old system, m=2 represents ages 4,5 and 6, etc.
	do age=1,pery
	m2(m)=m2(m)+(m1(m)+age-1)**2
	enddo
enddo


do m=1,ret_age-1
	eta(m)=coeff0+coeff1*real(m1(m),dp)+coeff2*real(m2(m),dp)/real(pery,dp)
	ybar(m)=0.0_dp
enddo
do m=ret_age, max_age   !in retirement, fixed endowment
	eta(m)=0.0_dp 
	ybar(m)=pension   
enddo


exp_eta=0.0_dp   !EXPECTED VALUE OF PERSISTENT PART (CHECK)

do m=1,ret_age-1

	exp_eta=exp_eta+eta(m)*p_age(m)

enddo

!re-scale so that mean of y is 1 exactly:
eta=eta/exp_eta

!re-check:
exp_eta=0.0_dp   !EXPECTED VALUE OF PERSISTENT PART (CHECK)

do m=1,ret_age-1

	exp_eta=exp_eta+eta(m)*p_age(m)

enddo

print*, "exp_eta=", exp_eta

open (unit=77774, file='eta.txt', status="replace")
write (77774, '(f16.8)') (eta)
close(unit=77774)


!*******************************************************************************
!DISCRETIZATION OF PERSISTENT PART OF INCOME
!*******************************************************************************


!discretize the state space:
!***************************

u_sigma=stdevy*dsqrt(1.0_dp-y_rho**2)   

ymax=width_y*stdevy
ymin=-ymax

w=(ymax-ymin)/real((nny-1),dp)

do iy1=1,nny

y(iy1)=ymin+real((iy1-1),dp)*w

enddo

open (unit = 131256, file = 'u_sigma.txt', status = 'replace')
write (131256,  '(f16.4)') (u_sigma)
close (unit=131256)


!calculate the transition matrix for idiosyncratic shock:
!*********************************************************


do iy1=1,nny

	pyy(iy1,1)=(DNORDF((y(1)-y_rho*y(iy1)+w/2.0_dp)/u_sigma))
	pyy(iy1,nny)=(1.0_dp-DNORDF((y(nny)-y_rho*y(iy1)-w/2.0_dp)/u_sigma))


	do iy2=2,nny-1

		 pyy(iy1,iy2)=(DNORDF((y(iy2)-y_rho*y(iy1)+w/2.0_dp)/u_sigma)-&
			DNORDF((y(iy2)-y_rho*y(iy1)-w/2.0_dp)/u_sigma))
		
	enddo  !end loop iy2
	

enddo  !end loop iy1


!check:

do iy1=1,nny

	if (sum(pyy(iy1,:))>=0.999_dp .and. sum(pyy(iy1,:))<=1.001_dp) then 

!	print*, "Markov approximation succeeded"
 
	else

    print*, "Markov approximation failed"
 
	endif

!print*, 'row=',iy1

enddo  !end iy1 loop


!print*, "transition matrix:", pyy

open (unit=7777, file='pyy.txt', status="replace")
do iy1=1,nny
do iy2=1,nny
write (7777, '(f16.8)') (pyy(iy1,iy2))
enddo
enddo 
close(unit=7777)

!********************************************
!calculate invariant distribution of
!Markov chain
!********************************************

transpyy=transpose(pyy)

py=(1.0_dp/real(nny,dp))*Ivect   !initial distribution of states

test=1.0_dp

do while (test>smalletto)

	py1=matmul(transpyy,py)

	test=maxval(dabs(py1-py))

	py=py1

enddo

print*, "py:", py

! Matteo: Modify income so that log income is normally distributed
! with distribution as given in mod_computation
! and mean income is given by mean_y


do iy1=1,nny

y(iy1)=dexp(y(iy1)-0.5_dp*stdevy**2.0_dp)   !+mean_y(gr)-1.0_dp

enddo


exp_y=0.0_dp   !EXPECTED VALUE OF PERSISTENT PART (CHECK)

do iy1=1,nny

	exp_y=exp_y+y(iy1)*py(iy1)

enddo

!re-scale so that mean of y is 1 exactly:
y=y/exp_y

!re-check:
exp_y=0.0_dp   !EXPECTED VALUE OF PERSISTENT PART (CHECK)

do iy1=1,nny

	exp_y=exp_y+y(iy1)*py(iy1)

enddo


var_y=0.0_dp   !VARIANCE OF PERSISTENT PART OF INCOME (CHECK)

do iy1=1,nny

var_y=var_y+py(iy1)*(dlog(y(iy1)))**2.0_dp

enddo

print*, "exp_y:", exp_y
print*, "var_y:", var_y

open (unit=7777, file='y.txt', status="replace")
write (7777, '(f16.4)') (y)
close(unit=7777)

open (unit=77775, file='py.txt', status="replace")
write (77775, '(f16.4)') (py)
close(unit=77775)


!*********************************************
!AGGREGATE SHOCK:
!*********************************************

!discretize the state space:
!***************************

u_sigmaa=stdevA*dsqrt(1.0_dp-rhoA**2)   

yamax=width_a*stdevA
yamin=-yamax

w=(yamax-yamin)/real((ns-1),dp)

do iy1=1,ns

ya(iy1)=yamin+real((iy1-1),dp)*w

enddo

open (unit = 131256, file = 'u_sigmaa.txt', status = 'replace')
write (131256,  '(f16.4)') (u_sigmaa)
close (unit=131256)


!calculate the transition matrix for idiosyncratic shock:
!*********************************************************


do iy1=1,ns

	pAA(iy1,1)=(DNORDF((ya(1)-rhoa*ya(iy1)+w/2.0_dp)/u_sigmaa))
	pAA(iy1,ns)=(1.0_dp-DNORDF((ya(ns)-rhoA*ya(iy1)-w/2.0_dp)/u_sigmaa))


	do iy2=2,ns-1

		 pAA(iy1,iy2)=(DNORDF((ya(iy2)-rhoA*ya(iy1)+w/2.0_dp)/u_sigmaa)-&
			DNORDF((ya(iy2)-rhoA*ya(iy1)-w/2.0_dp)/u_sigmaa))
		
	enddo  !end loop iy2
	

enddo  !end loop iy1


!check:

do iy1=1,ns

	if (sum(pAA(iy1,:))>=0.999_dp .and. sum(pAA(iy1,:))<=1.001_dp) then 

!	print*, "Markov approximation succeeded"
 
	else

    print*, "Markov approximation failed"
 
	endif

!print*, 'row=',iy1

enddo  !end iy1 loop


open (unit=7777, file='pAA.txt', status="replace")
do iy1=1,ns
do iy2=1,ns
write (7777, '(f16.8)') (pAA(iy1,iy2))
enddo
enddo 
close(unit=7777)

do iy1=1,ns

Ags(iy1)=dexp(ya(iy1)-0.5_dp*stdevA**2)  
logA(iy1)=dlog(Ags(iy1))

enddo

open (unit=7777, file='Ags.txt', status="replace")
write (7777, '(f16.4)') (Ags)
close(unit=7777)

!************************************************
!COMBINE WITH AGGREGATE SHOCK (stack everything)
!************************************************
m=1
do is=1,ns
n=1
do iss=1,ns
    payay((m-1)*nny+1:(m-1)*nny+nny,(n-1)*nny+1:(n-1)*nny+nny)=pAA(is,iss)*pyy(:,:)
    n=n+1
enddo
m=m+1
enddo

!payay=transpose(payay)


m=0
do is=1,ns
do iy=1,nny
    m=m+1
    Ay(m)=y(iy)
enddo
enddo
    

!check:

do iy1=1,ny

	if (sum(payay(iy1,:))>=0.999_dp .and. sum(payay(iy1,:))<=1.001_dp) then 

!	print*, "Big approximation succeeded"
 
	else

    print*, "Big approximation failed"
 
	endif

!print*, 'row=',iy1

enddo  !end iy1 loop

open (unit=7777, file='payay.txt', status="replace")
write (7777, '(f16.4)') (payay)
close(unit=7777)

!************************************************
!GENERATE FIXED EFFECT
!************************************************
!discretize the state space:
!***************************

if (nz==1) then
z(1)=mean_z
pz(1)=1.0_dp
else

zmax=width_z*stdevz
zmin=-zmax

we=(zmax-zmin)/real((nz-1),dp)

do iz=1,nz

z(iz)=zmin+real((iz-1),dp)*we

enddo


!calculate the fraction of agents for each z, if log(z) is normally distributed:
!*********************************************************************************

pz(1)=DNORDF((z(1)+we/2.0_dp)/stdevz)
pz(nz)=1.0_dp-DNORDF((z(nz)-we/2.0_dp)/stdevz)

do iz=2,nz-1

pz(iz)=DNORDF((z(iz)+we/2.0_dp)/stdevz)-DNORDF((z(iz)-we/2.0_dp)/stdevz)
		
enddo  !end loop iz


	!check:

	if (sum(pz)>=0.999_dp .and. sum(pz)<=1.001_dp) then 

	print*, "z approximation succeeded"
 
	else

    print*, "z approximation failed"
 
	endif

!z transformed:

do iz=1,nz

z(iz)=exp(z(iz)-0.5_dp*stdevz**2.0_dp) 

enddo
endif

open (unit=7777, file='z.txt', status="replace")
write (7777, '(f16.4)') (z)
close(unit=7777)
open (unit=7777, file='pz.txt', status="replace")
write (7777, '(f16.8)') (pz)
close(unit=7777)


!*********************************************
!BETAS:
!*********************************************

if (pbeta(1,1)<1.0_dp) then
!********************************************
!stationary distribution for beta:
!********************************************

transpb=transpose(pbeta)

pbetastat=(1.0_dp/real(nbeta,dp))*Ivect1   !initial distribution of states

test=1.0_dp

do while (test>smalletto)

	pb1=matmul(transpb,pbetastat)

	test=maxval(dabs(pb1-pbetastat))

	pbetastat=pb1

enddo

else
!***********************************************
!calibrated distribution of betas
!***********************************************

pbetastat=ppbeta

endif


open (unit=7777, file='pbeta.txt', status="replace")
write (7777, '(f16.4)') (pbeta)
close(unit=7777)

open (unit=77775, file='pbetastat.txt', status="replace")
write (77775, '(f16.4)') (pbetastat)
close(unit=77775)


!*****************************************************
!CONSTRUCT JOINT PROBABILITIES FOR Y AND BETAS:
!*****************************************************

 
m=1  !rows counter
do ib=1,nbeta
n=1 !columns counter
do ibprime=1,nbeta
	pbyy((m-1)*ny+1:(m-1)*ny+ny,(n-1)*ny+1:(n-1)*ny+ny)=pbeta(ib,ibprime)*payay(:,:)  !joint transition matrix
	n=n+1
enddo
m=m+1
enddo

!pbyy=transpose(pbyy)

m=1  !rows counter
do ib=1,nbeta
n=1 !columns counter
do ibprime=1,nbeta
	pbyy_sim((m-1)*nny+1:(m-1)*nny+nny,(n-1)*nny+1:(n-1)*nny+nny)=pbeta(ib,ibprime)*pyy(:,:)  !joint transition matrix
	n=n+1
enddo
m=m+1
enddo

!pbyy_sim=transpose(pbyy_sim)

!check:

do iy1=1,nyb

	if (sum(pbyy(iy1,:))>=0.999_dp .and. sum(pbyy(iy1,:))<=1.001_dp) then 

!	print*, "y-b trans approximation succeeded"
 
	else

    print*, "y-b trans approximation failed,sum=", sum(pbyy(iy1,:))
 
	endif

!print*, 'row=',iy1

enddo  !end iy1 loop

open (unit=7777, file='pbyy.txt', status="replace")
write (7777, '(8f16.8)') (transpose(pbyy))
close(unit=7777)


!*******************************************************************
!CONSTRUCT INCOME Z*Y and STATIONARY DISTRIBUTION FOR Z*Y AND BETA
!*******************************************************************

m=1
do iz=1,nz
do iy=1,ny
zy(m)=z(iz)*Ay(iy)
num(iz,iy)=m
m=m+1
enddo
enddo

open (unit=7777, file='zy.txt', status="replace")
write (7777, '(f16.4)') (zy)
close(unit=7777)


m=1
do iz=1,nz
do ib=1,nbeta
do is=1,ns
do iy=1,nny
pstat(m)=py(iy)*pbetastat(ib)*pz(iz)  !joint stationary distribution (at any given agg shock) 
index_beta(m)=ib	!index of beta corresponding to m
index_Azy(m)=num(iz,iy)  !index of zy corresponding to m
index_z(m)=iz
index_y(m)=iy
index_Ags(m)=is
m=m+1
enddo
enddo
enddo
enddo

!m=1
!do iz=1,nz
!do ib=1,nbeta
!do iy=1,nny
!pstat_sim(m)=py(iy)*pbetastat(ib)*pz(iz)  !joint stationary distribution (at any given agg shock) 
!m=m+1
!enddo
!enddo
!enddo

!**********************************************************************
!CONSTRUCT TRANSITION MATRIX TO BE USED IN COMPUTATION
!**********************************************************************

pbzyy=0.0_dp

do iz=1,nz

	do iy1=1,nyb

		do iy2=1,nyb

		pbzyy(nyb*(iz-1)+iy1,nyb*(iz-1)+iy2)=pbyy(iy1,iy2)

		enddo !end iy2

	enddo !end iy1

enddo  !end iz

!pbzyy_sim=0.0_dp

!do iz=1,nz

!	do iy1=1,nny*nbeta

!		do iy2=1,nny*nbeta

!		pbzyy_sim(nny*nbeta*(iz-1)+iy1,nny*nbeta*(iz-1)+iy2)=pbyy_sim(iy1,iy2)

!		enddo !end iy2

!	enddo !end iy1

!enddo  !end iz


open (unit=7777, file='pstat.txt', status="replace")
write (7777, '(f16.4)') (pstat)
close(unit=7777)

open (unit=77775, file='pbzyy.txt', status="replace")
write (77775, '(f16.4)') (pbzyy)
close(unit=77775)

!***********************************
!CONSTRUCT TOT INCOME in the economy
!***********************************

totlabor=0.0_dp
do age=1,max_age
	do iz=1,nz
	do iy=1,nny
	totlabor=totlabor+p_age(age)*pz(iz)*py(iy)*(y(iy)*z(iz)*eta(age))
	enddo
	enddo
enddo

!renormalize zy so that totlabor=1
do iz=1,ny*nz
zy(iz)=zy(iz)/totlabor
enddo


!just check for one aggregate shock (no difference for others)
totlabor=0.0_dp
do age=1,max_age
	do iz=1,nz
	do iy=1,nny
	totlabor=totlabor+p_age(age)*pz(iz)*py(iy)*(zy(iy)*eta(age))
	enddo
	enddo
!	print*, "totlabor:", totlabor
enddo

open (unit=77775, file='totlabor.txt', status="replace")
write (77775, '(f16.4)') (totlabor)
close(unit=77775)

end subroutine markovappr




end module mod_utils
